import pathlib
from typing import Any, Dict, List, Optional
from urllib.parse import urljoin

from boltons.setutils import IndexedSet
from telegraph import Telegraph
from tqdm import tqdm

from database.models import ExtractedPage
from database.repositories.extracted_page import ExtractedPageRepository
from grabber.core.utils import (
    build_unique_img_urls,
    get_all_posts_for_url,
    query_mapping,
    headers_mapping,
    get_tags,
    get_soup,
    run_downloader,
    send_post_to_telegram,
)


async def get_images_from_pagination(
    url: str,
    headers: Optional[dict] = None,
    query: Optional[str] = None,
) -> List[str]:
    page_nav_query = "div.page-links a.post-page-numbers"
    soup = await get_soup(url, headers=headers)
    page_counters = [int(page.get_text()) for page in soup.select(page_nav_query)]
    pages = [urljoin(url, f"{page_counter}/") for page_counter in page_counters]

    return pages


async def get_pages_from_pagination(
    url: str, headers: Optional[Dict[str, Any]] = None
) -> List[str]:
    pagination_pages_query = "div.page-nav.td-pb-padding-side a.page"
    posts_query = "div.td_block_inner.tdb-block-inner.td-fix-index div.td-module-meta-info > h3.entry-title.td-module-title > a"
    all_posts = set()
    soup = await get_soup(url, headers=headers)

    pages = {a.attrs["href"] for a in soup.select(pagination_pages_query)}
    pages.add(url)

    for page in pages:
        soup = await get_soup(page, headers=headers)
        all_posts.update({a.attrs["href"] for a in soup.select(posts_query)})

    return list(all_posts)


async def get_sources_for_spacemiss(
    sources: List[str],
    entity: str,
    telegraph_client: Optional[Telegraph] = None,
    final_dest: str | pathlib.Path = "",
    save_to_telegraph: bool | None = False,
    is_tag: Optional[bool] = False,
    limit: Optional[int] = None,
    **kwargs,
) -> None:
    query, src_attr = query_mapping[entity]
    headers = headers_mapping.get(entity, None)
    page_title = ""
    title_folder_mapping = {}
    posts_sent_counter = 0
    titles = IndexedSet()
    repository = ExtractedPageRepository(model=ExtractedPage)
    ordered_unique_img_urls = None
    all_sources = []

    if is_tag:
        sources = list(await get_pages_from_pagination(sources[0], headers=headers))

    tqdm_sources_iterable = tqdm(
        sources,
        total=len(sources),
        desc="Retrieving URLs...",
    )

    for source_url in tqdm_sources_iterable:
        all_sources += await get_all_posts_for_url(repository=repository, url=source_url)

        if source_url in all_sources:
            tqdm_sources_iterable.set_description(
                f"Skipping {source_url} since it was already posted"
            )
            continue

        image_tags, soup = await get_tags(source_url, headers=headers, query=query)
        folder_name = ""

        page_title = soup.find("title").text.split("-")[0].strip().rstrip()
        titles.add(page_title)

        ordered_unique_img_urls = await build_unique_img_urls(image_tags, src_attr)
        tqdm_sources_iterable.set_description(f"Finished retrieving images for {page_title}")

        if final_dest:
            final_dest = pathlib.Path(final_dest)
            if not final_dest.exists():
                final_dest.mkdir(parents=True, exist_ok=True)

            folder_name = page_title
            title_dest = final_dest / folder_name
            if not title_dest.exists():
                title_dest.mkdir(parents=True, exist_ok=True)
            title_folder_mapping[page_title] = (ordered_unique_img_urls, title_dest)

        if save_to_telegraph:
            all_sources = await send_post_to_telegram(
                ordered_unique_img_urls=ordered_unique_img_urls,
                page_title=page_title,
                telegraph_client=telegraph_client,
                posts_sent_counter=posts_sent_counter,
                tqdm_sources_iterable=tqdm_sources_iterable,
                all_sources=all_sources,
                source_url=source_url,
            )
        page_title = ""

    if final_dest and ordered_unique_img_urls:
        await run_downloader(
            final_dest=final_dest,
            page_title=page_title,
            unique_img_urls=ordered_unique_img_urls,
            titles=titles,
            title_folder_mapping=title_folder_mapping,
            headers=headers,
        )
